# Copyright (c) 2022 - 2024 Advanced Micro Devices, Inc. All rights reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os, sys, struct
import datetime, pytz
import nnir as ir

def generateLicenseForScript(f):
    f.write( \
"""################################################################################
#
# MIT License
#
# Copyright (c) 2022 - 2024 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
################################################################################

# This file is generated by nnir_to_zendnn.py on %s
""" % (datetime.datetime.now(tz=pytz.timezone('America/Los_Angeles')).isoformat()))

def generateLicenseForCPP(f):
        f.write( \
"""/*
#
# MIT License
#
# Copyright (c) 2022 - 2024 Advanced Micro Devices, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
*/

// This file is generated by nnir_to_zendnn.py on %s
""" % (datetime.datetime.now(tz=pytz.timezone('America/Los_Angeles')).isoformat()))

def generateCMakeFiles(graph,outputFolder):
    fileName = outputFolder + '/CMakeLists.txt'
    print('creating ' + fileName + ' ...')
    with open(fileName, 'w') as f:
        generateLicenseForScript(f)
        f.write( \
"""
cmake_minimum_required(VERSION 3.10)

# ROCM Path
if(DEFINED ENV{ROCM_PATH})
    set(ROCM_PATH $ENV{ROCM_PATH} CACHE PATH "Default ROCm installation path")
elseif(ROCM_PATH)
    message("-- INFO:ROCM_PATH Set -- ${ROCM_PATH}")
else()
    set(ROCM_PATH /opt/rocm CACHE PATH "Default ROCm installation path")
endif()
# Set AMD Clang as default compiler
if(NOT DEFINED CMAKE_CXX_COMPILER AND EXISTS "${ROCM_PATH}/lib/llvm/bin/amdclang++")
    set(CMAKE_CXX_COMPILER ${ROCM_PATH}/lib/llvm/bin/amdclang++)
endif()

# Set Default Compiler & Standard
# aocc-linux-compiler -- ENV:ZENDNN_AOCC_COMP_PATH
if(NOT DEFINED ZENDNN_AOCC_COMP_PATH)
    if(NOT DEFINED ENV{ZENDNN_AOCC_COMP_PATH})
        message(FATAL_ERROR "ZENDNN_AOCC_COMP_PATH NOT FOUND -- aocc-compiler-X.X.XX install path must be set to ZENDNN_AOCC_COMP_PATH${ColourReset}")
    else()
        set(ZENDNN_AOCC_COMP_PATH $ENV{ZENDNN_AOCC_COMP_PATH} CACHE PATH "Path to which ZENDNN_AOCC_COMP_PATH has been installed")
    endif()
endif()
set(CMAKE_CXX_STANDARD 17)
set(CMAKE_CXX_STANDARD_REQUIRED On)
set(CMAKE_CXX_EXTENSIONS ON)
set(CMAKE_C_COMPILER ${ZENDNN_AOCC_COMP_PATH}/bin/clang)
set(CMAKE_CXX_COMPILER ${ZENDNN_AOCC_COMP_PATH}/bin/clang++)

project(zendnn_app)

# OpenMP
find_package(OpenMP REQUIRED)
set(CMAKE_C_FLAGS "${CMAKE_C_FLAGS} ${OpenMP_C_FLAGS}")
# -fopenmp -- Enable OpenMP
set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} ${OpenMP_CXX_FLAGS}")
set(LINK_LIBRARY_LIST ${LINK_LIBRARY_LIST} OpenMP::OpenMP_CXX)

# aocl-linux-aocc -- ENV:ZENDNN_BLIS_PATH
if(NOT DEFINED ZENDNN_BLIS_PATH)
    if(NOT DEFINED ENV{ZENDNN_BLIS_PATH})
        message(FATAL_ERROR "ZENDNN_BLIS_PATH NOT FOUND -- aocl-linux-aocc-X.X.XX install path must be set to ZENDNN_BLIS_PATH${ColourReset}")
    else()
        set(ZENDNN_BLIS_PATH $ENV{ZENDNN_BLIS_PATH} CACHE PATH "Path to which ZENDNN_BLIS_PATH has been installed")
    endif()
endif()

# amd Zen DNN Git Source
if(NOT DEFINED ZENDNN_GIT_ROOT)
    if(NOT DEFINED ENV{ZENDNN_GIT_ROOT})
        message(FATAL_ERROR "ZENDNN_GIT_ROOT NOT FOUND -- ZEN DNN Git Source path must be set to ZENDNN_GIT_ROOT${ColourReset}")
    else()
        set(ZENDNN_GIT_ROOT $ENV{ZENDNN_GIT_ROOT} CACHE PATH "Path to which ZENDNN_GIT_ROOT has been installed")
    endif()
endif()

# set global compiler flags for the project
# -O3 -- Optimize output file
# -fPIC -- Generate position-independent code if possible.
# -march -- Generate code for given CPU. [znver2 -- Zen Version 2]
# -Wreturn-type -- Warn whenever a function's return type defaults to "int" (C), or about inconsistent return types (C++).
set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} -DLIBM_ENABLE=1 -DBIAS_ENABLED=1 -DZENDNN_ENABLE=1 -DZENDNN_X64=1")
set(CMAKE_CXX_FLAGS "${CMAKE_CXX_FLAGS} -O3 -fPIC -march=znver2 -Wreturn-type")

# INCLUDE FILES
# ZENDNN_GIT_ROOT includes
include_directories(${ZENDNN_GIT_ROOT}/inc ${ZENDNN_GIT_ROOT}/src ${ZENDNN_GIT_ROOT}/src/common ${ZENDNN_GIT_ROOT}/src/cpu)
# ZENDNN_BLIS_PATH includes
include_directories(${ZENDNN_BLIS_PATH}/include)
# GNU Includes
include_directories(/usr/lib/x86_64-linux-gnu/include)
# test utilities
include_directories(${ZENDNN_GIT_ROOT}/tests/api_tests)

# Link Library Directories
link_directories(${ZENDNN_GIT_ROOT}/_out/lib)
link_directories(${ZENDNN_BLIS_PATH}/lib)
link_directories(/usr/lib/x86_64-linux-gnu/lib)

file(GLOB My_Source_Files ./source/*.cpp)
add_executable(${PROJECT_NAME} ${My_Source_Files})

set(LINK_LIBRARY_LIST ${LINK_LIBRARY_LIST} amdZenDNN blis-mt alm)

# OpenCV -- Display Component
find_package(OpenCV REQUIRED)
# OpenCV 3/4 Support
if(${OpenCV_VERSION_MAJOR} EQUAL 3 OR ${OpenCV_VERSION_MAJOR} EQUAL 4)
	message("-- ${PROJECT_NAME} -- OpenCV Version Supported -- Version-${OpenCV_VERSION_MAJOR}.${OpenCV_VERSION_MINOR}.X")
	if(${OpenCV_VERSION_MAJOR} EQUAL 4)
	    target_compile_definitions(${PROJECT_NAME} PUBLIC USE_OPENCV_4=1)
    else()
	    target_compile_definitions(${PROJECT_NAME} PUBLIC USE_OPENCV_4=0)
    endif()
	include_directories(${OpenCV_INCLUDE_DIRS})
    #set(LINK_LIBRARY_LIST ${LINK_LIBRARY_LIST} ${OpenCV_LIBRARIES}) # TBD -- Investigate SegFault
else()
	message(FATAL_ERROR "${PROJECT_NAME} -- OpenCV -- Version-${OpenCV_VERSION_MAJOR}.${OpenCV_VERSION_MINOR}.X Not Supported")
endif()

target_link_libraries(${PROJECT_NAME} ${LINK_LIBRARY_LIST})

# Bring in weights & images to build folder
FILE(GLOB IMAGE_LIST ${CMAKE_CURRENT_SOURCE_DIR}/data/images/*.png)
FILE(GLOB IMAGE_BIN_LIST ${CMAKE_CURRENT_SOURCE_DIR}/data/images/*.bin)
file(COPY ${CMAKE_CURRENT_SOURCE_DIR}/data/weights.bin DESTINATION ${CMAKE_CURRENT_BINARY_DIR}/data/)
file(COPY ${IMAGE_LIST} DESTINATION ${CMAKE_CURRENT_BINARY_DIR}/images/)
file(COPY ${IMAGE_BIN_LIST} DESTINATION ${CMAKE_CURRENT_BINARY_DIR}/images/)

message("-- ${PROJECT_NAME} -- Using Compiler - Path:" ${CMAKE_CXX_COMPILER} "\t
        Version:" ${CMAKE_CXX_COMPILER_VERSION} "\tCompiler:" ${CMAKE_CXX_COMPILER_ID}${ColourReset})
message("-- ${PROJECT_NAME} -- CMAKE_CXX_FLAGS:${CMAKE_CXX_FLAGS}${ColourReset}")
message("-- ${PROJECT_NAME} -- Link Libraries: ${LINK_LIBRARY_LIST}${ColourReset}")
""")

def generateDeployCPP(graph,fileName):
    print('creating ' + fileName + ' ...')
    with open(fileName, 'w') as f:
        generateLicenseForCPP(f)
        f.write( \
"""
#include <assert.h>

#include <chrono>
#include <numeric>
#include <vector>
#include <unordered_map>
#include <iostream>
#include <fstream>
#include <algorithm>

#include "test_utils.hpp"
#include "zendnn_logging.hpp"

#include <opencv2/opencv.hpp>

using namespace zendnn;
using namespace std;

#if USE_OPENCV_4
#define CV_BGR2GRAY COLOR_BGR2GRAY
#endif

memory::dim product(const memory::dims &dims)
{
    return std::accumulate(dims.begin(), dims.end(), (memory::dim)1,
                           std::multiplies<memory::dim>());
}

// Check function status
#define ERROR_CHECK_STATUS(call)                                                               \
    {                                                                                          \
        int status = (call);                                                                   \
        if (status != 0)                                                                       \
        {                                                                                      \
            printf("ERROR: failed with status = (%d) at " __FILE__ "#%d\\n", status, __LINE__); \
            return status;                                                                     \
        }                                                                                      \
    }

// Intialize weights & bias tensors
static int initializeTensor(std::vector<float> *tensor, size_t tensorSize, FILE *fp, const char *binaryFilename)
{
    size_t itemsize = sizeof(float);
    unsigned int h[2] = {0};

    fread(h, 1, sizeof(h), fp);
    if (h[0] != 0xf00dd1e1 || (size_t)h[1] != (tensorSize * itemsize))
    {
        printf("ERROR: invalid data (magic,size)=(0x%x,%x) in %s at byte position %lu -- expected size is %ld\\n", 
                h[0], h[1], binaryFilename, ftell(fp) - sizeof(h), tensorSize * itemsize);
        return -1;
    }

    void *ptr = tensor->data();
    size_t n = fread(ptr, itemsize, tensorSize, fp);
    if (n != tensorSize)
    {
        printf("ERROR: expected char[%zu], but got char[%zu] in %s\\n", tensorSize * itemsize, n * itemsize, binaryFilename);
        return -1;
    }

    return 0;
}

// Model Zen DNN Implementation
int model_setup(engine::kind engine_kind, const char *binaryFilename, const char *imageFilename, int times)
{

    using tag = memory::format_tag;
    using dt = memory::data_type;

    // Check weights file
    FILE *fp__variables = fopen(binaryFilename, "rb");
    if (!fp__variables)
    {
        printf("ERROR: unable to open: %s\\n", binaryFilename);
        return -1;
    }
    {
        unsigned int magic = 0;
        fread(&magic, 1, sizeof(magic), fp__variables);
        if (magic != 0xf00dd1e0)
        {
            fclose(fp__variables);
            printf("ERROR: invalid file magic in %s\\n", binaryFilename);
            return -1;
        }
    }

    // Initialize engine and stream
    engine eng(engine_kind, 0);
    stream s(eng);

    // Create network
    std::vector<primitive> net;
    std::vector<std::unordered_map<int, memory>> net_args;
""")
        for tensor in graph.inputs:
            f.write( \
"""
    // Allocate input data %s
    std::vector<float> user_src(%s);
    memory::dims %s_src_tensor_dims = {%s};

    //Create memory that describes data layout
    auto user_src_memory = memory({{%s_src_tensor_dims}, dt::f32, tag::nchw}, eng);
    write_to_zendnn_memory(user_src.data(), user_src_memory);
"""% (tensor.name, ' * '.join([str(v) for v in tensor.shape]), 
        tensor.name, ', '.join([str(v) for v in tensor.shape]), tensor.name))
        for tensor in graph.outputs:
            f.write( \
"""
    // Allocate output data %s
    std::vector<float> user_dst(%s);
    memory::dims %s_dims = {%s};
"""% (tensor.name, ' * '.join([str(v) for v in tensor.shape]), tensor.name, ', '.join([str(v) for v in tensor.shape])))
        for tensor in graph.inputs:
            f.write( \
"""
    // Load Input Image - Binary Files
    std::ifstream input_binary_image(imageFilename, std::ios::binary);

    input_binary_image.seekg(0, input_binary_image.end);
    int elementCount = (input_binary_image.tellg() / sizeof(float)); // total number of elements
    input_binary_image.seekg(0, input_binary_image.beg);

    if (elementCount == (%s))
    {
        input_binary_image.read(reinterpret_cast<char *>(user_src.data()), user_src.size() * sizeof(float));
    }
    else
    {
        printf("ERROR: invalid Binary Image File -- %%s: Total Pixels:%s Received: %%d\\n", imageFilename, elementCount);
        return -1;
    }

    // Allocate initializers
"""% (' * '.join([str(v) for v in tensor.shape]), ' * '.join([str(v) for v in tensor.shape])))
        for tensor in graph.initializers:
            i = 0
            while i < len(tensor.shape):
                if tensor.shape[-1] == 1:
                    del tensor.shape[-1]
                i += 1
            f.write( \
"""    memory::dims %s_dims = {%s};
"""%(tensor.name, ', '.join([str(v) for v in tensor.shape])))
        f.write( 
"""
    // Allocate locals
""")
        for tensor in graph.locals:
            j = 0
            while j < len(tensor.shape):
                if tensor.shape[-1] == 1:
                    del tensor.shape[-1]
                j += 1
            f.write( \
"""    memory::dims %s_dims = {%s};
"""%(tensor.name, ', '.join([str(v) for v in tensor.shape])))
        layerNumber = 1
        for node in graph.nodes:
            if node.type == 'conv':
                # TODO: add support for group convolution
                pads = node.attr.get('pads')
                strides = node.attr.get('strides')
                f.write( \
"""
    // Start: Model Layer %d - conv
    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - conv Setup");
    // Conv set dimensions
    memory::dims %s_padding = {%s};
    memory::dims %s_strides = {%s};
    // Allocate buffers for weights and bias
    std::vector<float> %s(product(%s_dims));
    ERROR_CHECK_STATUS(initializeTensor(&%s, product(%s_dims), fp__variables, binaryFilename));
"""%(layerNumber, layerNumber, node.outputs[0], ', '.join([str(v) for v in pads]), 
    node.outputs[0], ', '.join([str(v) for v in strides]), node.inputs[1], node.inputs[1], 
    node.inputs[1], node.inputs[1]))
                if len(node.inputs) == 3:
                    f.write( \
"""
    std::vector<float> %s(product(%s_dims));
    ERROR_CHECK_STATUS(initializeTensor(&%s, product(%s_dims), fp__variables, binaryFilename));
"""%(node.inputs[2], node.inputs[2], node.inputs[2], node.inputs[2]))
                f.write( \
"""
    // Create memory that describes data layout in the buffers
    auto %s_user_weights_memory = memory({{%s_dims}, dt::f32, tag::oihw}, eng);
    write_to_zendnn_memory(%s.data(), %s_user_weights_memory);
"""%(node.outputs[0], node.inputs[1], node.inputs[1], node.outputs[0]))
                if len(node.inputs) == 3:
                    f.write( \
"""
    auto %s_user_bias_memory = memory({{%s_dims[1]}, dt::f32, tag::x}, eng);
    write_to_zendnn_memory(%s.data(), %s_user_bias_memory);

    // Create convolution memory descriptors with layout tag::any
"""%(node.outputs[0], node.inputs[2], node.inputs[2], node.outputs[0]))
                if layerNumber == 1:
                    f.write( \
"""
    auto %s_src_md = memory::desc({%s_src_tensor_dims}, dt::f32, tag::any);
"""%(node.outputs[0], node.inputs[0]))
                if len(node.inputs) == 3:
                    f.write( \
"""
    auto %s_bias_md = memory::desc({%s_dims[1]}, dt::f32, tag::any);
"""%(node.outputs[0], node.inputs[2]))
                f.write( \
"""
    auto %s_weights_md = memory::desc({%s_dims}, dt::f32, tag::any);
    auto %s_dst_md = memory::desc({%s_dims}, dt::f32, tag::any);

    // Create convolution descriptor
"""%(node.outputs[0], node.inputs[1], node.outputs[0], node.outputs[0]))
                if layerNumber == 1:
                    if len(node.inputs) == 3:
                        f.write( \
"""
    auto %s_desc = convolution_forward::desc(prop_kind::forward_inference,
                                            algorithm::convolution_gemm, %s_src_md, %s_weights_md,
                                            %s_bias_md, %s_dst_md, %s_strides, %s_padding,
                                            %s_padding);
"""%(node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0],
    node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0]))
                    else:
                        f.write( \
"""
    auto %s_desc = convolution_forward::desc(prop_kind::forward_inference,
                                            algorithm::convolution_gemm, %s_src_md, %s_weights_md,
                                            %s_dst_md, %s_strides, %s_padding, %s_padding);
"""%(node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0],
    node.outputs[0], node.outputs[0], node.outputs[0]))
                else:
                    if len(node.inputs) == 3:
                        f.write( \
"""
    auto %s_desc = convolution_forward::desc(prop_kind::forward_inference,
                                            algorithm::convolution_gemm, %s_dst_memory.get_desc(), %s_weights_md,
                                            %s_bias_md, %s_dst_md, %s_strides, %s_padding,
                                            %s_padding);
    // End: Create convolution descriptor
"""%(node.outputs[0], node.inputs[0], node.outputs[0], node.outputs[0],
    node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0]))
                    else:
                        f.write( \
"""
    auto %s_desc = convolution_forward::desc(prop_kind::forward_inference,
                                            algorithm::convolution_gemm, %s_dst_memory.get_desc(), %s_weights_md,
                                            %s_dst_md, %s_strides, %s_padding, %s_padding);
    // End: Create convolution descriptor
"""%(node.outputs[0], node.inputs[0], node.outputs[0], node.outputs[0],
    node.outputs[0], node.outputs[0], node.outputs[0]))
                f.write( \
"""
    // Create a convolution primitive descriptor
    auto %s_prim_desc = convolution_forward::primitive_desc(%s_desc, eng);
"""%(node.outputs[0], node.outputs[0]))
                if layerNumber == 1:
                    f.write( \
"""
    // Check data and weights formats - Reorder
    auto %s_src_memory = user_src_memory;
    if (%s_prim_desc.src_desc() != user_src_memory.get_desc())
    {
        %s_src_memory = memory(%s_prim_desc.src_desc(), eng);
        net.push_back(reorder(user_src_memory, %s_src_memory));
        net_args.push_back({{ZENDNN_ARG_FROM, user_src_memory},
                            {ZENDNN_ARG_TO, %s_src_memory}});
    }

    auto %s_weights_memory = %s_user_weights_memory;
    if (%s_prim_desc.weights_desc() != %s_user_weights_memory.get_desc())
    {
        %s_weights_memory = memory(%s_prim_desc.weights_desc(), eng);
        reorder(%s_user_weights_memory, %s_weights_memory)
            .execute(s, %s_user_weights_memory, %s_weights_memory);
    }
"""%(node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0],
    node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0],
    node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0],
    node.outputs[0]))
                else:
                    f.write( \
"""
    // Check weights formats - Reorder
    auto %s_weights_memory = %s_user_weights_memory;
    if (%s_prim_desc.weights_desc() != %s_user_weights_memory.get_desc())
    {
        %s_weights_memory = memory(%s_prim_desc.weights_desc(), eng);
        reorder(%s_user_weights_memory, %s_weights_memory)
            .execute(s, %s_user_weights_memory, %s_weights_memory);
    }
"""%(node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0],
    node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0]))
                f.write( \
"""
    // Create a memory primitive for output
    auto %s_dst_memory = memory(%s_prim_desc.dst_desc(), eng);
"""%(node.outputs[0], node.outputs[0]))
                if layerNumber == 1:
                    if len(node.inputs) == 3:
                        f.write( \
"""
    // Create a convolution primitive and add it to the net
    net.push_back(convolution_forward(%s_prim_desc));
    net_args.push_back({{ZENDNN_ARG_SRC, %s_src_memory},
                        {ZENDNN_ARG_WEIGHTS, %s_weights_memory},
                        {ZENDNN_ARG_BIAS, %s_user_bias_memory},
                        {ZENDNN_ARG_DST, %s_dst_memory}});
"""%(node.outputs[0], node.outputs[0], node.outputs[0],
    node.outputs[0], node.outputs[0]))
                    else:
                        f.write( \
"""
    // Create a convolution primitive and add it to the net
    net.push_back(convolution_forward(%s_prim_desc));
    net_args.push_back({{ZENDNN_ARG_SRC, %s_src_memory},
                        {ZENDNN_ARG_WEIGHTS, %s_weights_memory},
                        {ZENDNN_ARG_DST, %s_dst_memory}});
"""%(node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0]))
                else:
                    if len(node.inputs) == 3:
                        f.write( \
"""
    // Create a convolution primitive and add it to the net
    net.push_back(convolution_forward(%s_prim_desc));
    net_args.push_back({{ZENDNN_ARG_SRC, %s_dst_memory},
                        {ZENDNN_ARG_WEIGHTS, %s_weights_memory},
                        {ZENDNN_ARG_BIAS, %s_user_bias_memory},
                        {ZENDNN_ARG_DST, %s_dst_memory}});
"""%(node.outputs[0], node.inputs[0], node.outputs[0],
    node.outputs[0], node.outputs[0]))
                    else:
                        f.write( \
"""
    // Create a convolution primitive and add it to the net
    net.push_back(convolution_forward(%s_prim_desc));
    net_args.push_back({{ZENDNN_ARG_SRC, %s_dst_memory},
                        {ZENDNN_ARG_WEIGHTS, %s_weights_memory},
                        {ZENDNN_ARG_DST, %s_dst_memory}});
"""%(node.outputs[0], node.inputs[0], node.outputs[0], node.outputs[0]))
                f.write( \
"""
    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - conv Setup Complete");
    // End: Model Layer %d - conv
"""%(layerNumber, layerNumber))
                layerNumber += 1
            elif node.type == 'max_pool' or node.type == 'avg_pool':
                pads = node.attr.get('pads')
                strides = node.attr.get('strides')
                kernel = node.attr.get('kernel_shape')
                f.write( \
"""
    // Start: Model Layer %d - pool
    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - pool Setup");
    // Pool set dimensions
    memory::dims %s_padding = {%s};
    memory::dims %s_strides = {%s};
    memory::dims %s_kernel = {%s};

    // Pool - set dst memory desc
    auto %s_dst_md = memory::desc({%s_dims}, dt::f32, tag::any);

    // Create pooling primitive
    auto %s_desc = pooling_forward::desc(prop_kind::forward_inference,
                                            algorithm::%s, %s_dst_memory.get_desc(), %s_dst_md,
                                            %s_strides, %s_kernel, %s_padding, %s_padding);
    auto %s_pd = pooling_forward::primitive_desc(%s_desc, eng);

    // Pool - dst memory
    auto %s_dst_memory = memory(%s_pd.dst_desc(), eng);

    // Create a pooling primitive and add it to the net
    net.push_back(pooling_forward(%s_pd));
    net_args.push_back({{ZENDNN_ARG_SRC, %s_dst_memory},
                        {ZENDNN_ARG_DST, %s_dst_memory}});
    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - pool Setup Complete");
    // End: Model Layer %d - pool

"""%(layerNumber, layerNumber, node.outputs[0], ', '.join([str(v) for v in pads]), node.outputs[0], 
    ', '.join([str(v) for v in strides]), node.outputs[0], ', '.join([str(v) for v in kernel]), node.outputs[0], 
    node.outputs[0], node.outputs[0], "pooling_max" if node.type == 'max_pool' else "pooling_avg",
    node.inputs[0], node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], 
    node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], 
    node.outputs[0], node.inputs[0], node.outputs[0], layerNumber, layerNumber))
                layerNumber += 1
            elif node.type == 'gemm':
                weights_shape = graph.tensor_dict[node.inputs[1]].shape
                f.write( \
"""
    // Start: Model Layer %d - inner product
    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - inner product Setup");

    // Weights and bias
    std::vector<float> %s_weights(product(%s_dims));
    ERROR_CHECK_STATUS(initializeTensor(&%s_weights, product(%s_dims), fp__variables, binaryFilename));
    std::vector<float> %s_bias(product(%s_dims));
    ERROR_CHECK_STATUS(initializeTensor(&%s_bias, product(%s_dims), fp__variables, binaryFilename));

    // Create memory that describes data layout in the buffers
    auto %s_user_weights_memory = memory({{%s_dims}, dt::f32, tag::%s}, eng);
    write_to_zendnn_memory(%s_weights.data(), %s_user_weights_memory);
    auto %s_user_bias_memory = memory({{%s_dims[1]}, dt::f32, tag::x}, eng);
    write_to_zendnn_memory(%s_bias.data(), %s_user_bias_memory);

    // Create ip memory descriptors with layout tag::any
    auto %s_weights_md = memory::desc({%s_dims}, dt::f32, tag::any);
    auto %s_bias_md = memory::desc({%s_dims[1]}, dt::f32, tag::any);
    auto %s_dst_md = memory::desc({%s_dims}, dt::f32, tag::any);

    // Create inner product primitive descriptor
    auto %s_desc = inner_product_forward::desc(prop_kind::forward_inference,
                                                %s_dst_memory.get_desc(), %s_weights_md, %s_bias_md, 
                                                %s_dst_md);
    auto %s_prim_desc = inner_product_forward::primitive_desc(%s_desc, eng);
"""%(layerNumber, layerNumber, node.inputs[1], node.inputs[1], node.inputs[1], 
    node.inputs[1], node.inputs[2], node.inputs[2], node.inputs[2], node.inputs[2], 
    node.inputs[1], node.inputs[1], 'oihw' if len(weights_shape) == 4 else 'nc', 
    node.inputs[1], node.inputs[1], node.inputs[2],
    node.inputs[2], node.inputs[2], node.inputs[2], node.inputs[1], node.inputs[1], 
    node.inputs[2], node.inputs[2], node.outputs[0], node.outputs[0], node.outputs[0], 
    node.inputs[0], node.inputs[1], node.inputs[2], node.outputs[0], node.outputs[0], 
    node.outputs[0]))
                f.write( \
"""
    // Check weights formats - Reorder
    auto %s_weights_memory = %s_user_weights_memory;
    if (%s_prim_desc.weights_desc() != %s_user_weights_memory.get_desc())
    {
        %s_weights_memory = memory(%s_prim_desc.weights_desc(), eng);
        reorder(%s_user_weights_memory, %s_weights_memory)
            .execute(s, %s_user_weights_memory, %s_weights_memory);
    }

    // IP - dst memory
    auto %s_dst_memory = memory(%s_prim_desc.dst_desc(), eng);

    // Create a IP primitive and add it to the net
    net.push_back(inner_product_forward(%s_prim_desc));
    net_args.push_back({{ZENDNN_ARG_SRC, %s_dst_memory},
                        {ZENDNN_ARG_WEIGHTS, %s_weights_memory},
                        {ZENDNN_ARG_BIAS, %s_user_bias_memory},
                        {ZENDNN_ARG_DST, %s_dst_memory}});

    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - inner product Setup Complete");
    // End: Model Layer %d - inner product
"""%(node.inputs[1], node.inputs[1], node.outputs[0], node.inputs[1], node.inputs[1],
    node.outputs[0], node.inputs[1], node.inputs[1], node.inputs[1], node.inputs[1],
    node.outputs[0], node.outputs[0], node.outputs[0], node.inputs[0], node.inputs[1], 
    node.inputs[2], node.outputs[0], layerNumber, layerNumber))
                layerNumber += 1
            elif node.type == 'relu':
                alpha = node.attr.get('alpha')
                f.write( \
"""
    // Start: Model Layer %d - relu
    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - relu Setup");
    const float %s_alpha = %f;
    const float %s_beta = 0.0f;

    auto %s_dst_md = memory::desc({%s_dims}, dt::f32, tag::any);

    // Create relu primitive descriptor
    auto %s_desc = eltwise_forward::desc(prop_kind::forward_inference,
                                            algorithm::eltwise_relu, %s_dst_memory.get_desc(),
                                            %s_alpha, %s_beta);
    auto %s_prim_desc = eltwise_forward::primitive_desc(%s_desc, eng);

    // Relu - dst memory
    auto %s_dst_memory = memory(%s_prim_desc.dst_desc(), eng);

    // Create a relu primitive and add it to the net
    net.push_back(eltwise_forward(%s_prim_desc));
    net_args.push_back({{ZENDNN_ARG_SRC, %s_dst_memory},
                        {ZENDNN_ARG_DST, %s_dst_memory}});
    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - relu Setup Complete");
    // End: Model Layer %d - relu
"""%(layerNumber, layerNumber, node.outputs[0], alpha, node.outputs[0], node.outputs[0], 
    node.outputs[0],node.outputs[0], node.inputs[0], node.outputs[0], node.outputs[0], 
    node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], 
    node.inputs[0], node.outputs[0], layerNumber, layerNumber))
                layerNumber += 1
            elif node.type == 'lrn':
                size = node.attr.get('size')
                alpha = node.attr.get('alpha')
                beta = node.attr.get('beta')
                bias = node.attr.get('bias')
                f.write( \
"""
    // Start: Model Layer %d - LRN
    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - LRN Setup");
    const float %s_alpha = %f;
    const float %s_beta = %f;
    const float %s_bias = %f;
    const unsigned long %s_size = %lu;

    // Create LRN primitive descriptor
    auto %s_desc = lrn_forward::desc(prop_kind::forward_inference,
                                    algorithm::%s, 
                                    %s_dst_memory.get_desc(), %s_size, %s_alpha, %s_beta, %s_bias);

    auto %s_prim_desc = lrn_forward::primitive_desc(%s_desc, eng);

    // LRN - dst memory
    auto %s_dst_memory = memory(%s_prim_desc.dst_desc(), eng);

    // Create a LRN primitive and add it to the net
    net.push_back(lrn_forward(%s_prim_desc));
    net_args.push_back({{ZENDNN_ARG_SRC, %s_dst_memory},
                        {ZENDNN_ARG_DST, %s_dst_memory}});
    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - LRN Setup Complete");
    // End: Model Layer %d - LRN
"""%(layerNumber, layerNumber, node.outputs[0], alpha, node.outputs[0], beta, node.outputs[0], 
    bias, node.outputs[0], size, node.outputs[0] , 
    "lrn_across_channels" if node.attr.get('mode') == 0 else "lrn_within_channel",
    node.inputs[0], node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0],
    node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0],
    node.inputs[0], node.outputs[0], layerNumber, layerNumber))
                layerNumber += 1
            elif node.type == 'batch_norm':
                epsilon = node.attr.get('epsilon')
                f.write( \
"""
    // Start: Model Layer %d - Batch Normalization
    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - Batch Normalization Setup");
    const float %s_epsilon = %ef;

    // Allocate buffers for mean, variance, scale and bias
    std::vector<float> %s(product(%s_dims));
    ERROR_CHECK_STATUS(initializeTensor(&%s, product(%s_dims), fp__variables, binaryFilename));
    std::vector<float> %s(product(%s_dims));
    ERROR_CHECK_STATUS(initializeTensor(&%s, product(%s_dims), fp__variables, binaryFilename));
    std::vector<float> %s(product(%s_dims));
    ERROR_CHECK_STATUS(initializeTensor(&%s, product(%s_dims), fp__variables, binaryFilename));
    std::vector<float> %s(product(%s_dims));
    ERROR_CHECK_STATUS(initializeTensor(&%s, product(%s_dims), fp__variables, binaryFilename));

    // Allocate memory for scaleshift tensor
    memory::dims %s_scale_shift_dims = {2, %s};    
    std::vector<float> %s_scale_shift_data(product(%s_scale_shift_dims));

    auto %s_mid = %s_scale_shift_data.begin() + %s_scale_shift_dims[1];
    
    copy(%s.begin(), %s.end(), %s_scale_shift_data.begin());
    copy(%s.begin(), %s.end(), %s_mid);

    // Create memory that describes data layout in the buffers
    auto %s_user_mean_memory = memory({{%s_dims}, dt::f32, tag::nc}, eng);
    write_to_zendnn_memory(%s.data(), %s_user_mean_memory);
    auto %s_user_variance_memory = memory({{%s_dims}, dt::f32, tag::nc}, eng);
    write_to_zendnn_memory(%s.data(), %s_user_variance_memory);
    auto %s_user_scale_memory = memory({{%s_dims}, dt::f32, tag::nc}, eng);
    write_to_zendnn_memory(%s.data(), %s_user_scale_memory);
    auto %s_user_bias_memory = memory({{%s_dims}, dt::f32, tag::nc}, eng);
    write_to_zendnn_memory(%s.data(), %s_user_bias_memory);
    auto %s_scale_shift_memory = memory({{%s_scale_shift_dims}, dt::f32, tag::nc}, eng);
    write_to_zendnn_memory(%s_scale_shift_data.data(), %s_scale_shift_memory);

    // Create memory descriptors and memory objects. 
    auto %s_dst_md = memory::desc({%s_dims}, dt::f32, tag::any);

    auto %s_desc = batch_normalization_forward::desc(
            prop_kind::forward_inference, %s_dst_memory.get_desc(), %s_epsilon,
            normalization_flags::use_scale_shift);

    // Create primitive descriptor.
    auto %s_prim_desc = batch_normalization_forward::primitive_desc(%s_desc, eng);

    // Create a memory primitive for output
    auto %s_dst_memory = memory(%s_prim_desc.dst_desc(), eng);

    // Create a batch normalization primitive and add it to the net
    net.push_back(batch_normalization_forward(%s_prim_desc));
    net_args.push_back({{ZENDNN_ARG_SRC, %s_dst_memory},
                        {ZENDNN_ARG_MEAN, %s_user_mean_memory},
                        {ZENDNN_ARG_VARIANCE, %s_user_variance_memory},
                        {ZENDNN_ARG_SCALE_SHIFT, %s_scale_shift_memory},
                        {ZENDNN_ARG_DST, %s_dst_memory}});

    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - Batch Normalization Setup Complete");
    // End: Model Layer %d - Batch Normalization
"""%(layerNumber, layerNumber, node.outputs[0], epsilon, 
    node.inputs[1], node.inputs[1], node.inputs[1], node.inputs[1],
    node.inputs[2], node.inputs[2], node.inputs[2], node.inputs[2],
    node.inputs[3], node.inputs[3], node.inputs[3], node.inputs[3],
    node.inputs[4], node.inputs[4], node.inputs[4], node.inputs[4],
    node.outputs[0], str(graph.tensor_dict[node.inputs[0]].shape[1]), 
    node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], 
    node.outputs[0], node.inputs[1], node.inputs[1], node.outputs[0], 
    node.inputs[2], node.inputs[2], node.outputs[0], node.inputs[3], 
    node.inputs[3], node.inputs[3], node.inputs[3], node.inputs[4], 
    node.inputs[4], node.inputs[4], node.inputs[4], node.inputs[1], 
    node.inputs[1], node.inputs[1], node.inputs[1], node.inputs[2], 
    node.inputs[2], node.inputs[2], node.inputs[2], node.outputs[0], 
    node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], 
    node.outputs[0], node.outputs[0], node.inputs[0], node.outputs[0], 
    node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], 
    node.outputs[0], node.inputs[0], node.inputs[3],node.inputs[4], 
    node.outputs[0], node.outputs[0], layerNumber, layerNumber))
                layerNumber += 1
            elif node.type == 'sum' or node.type == 'add':
                num_src = len(node.inputs)
                f.write( \
"""
    // Start: Model Layer %d - Sum
    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - Sum Setup");

    // Sum number of inputs.
    const int %s_num_src = %d;

    // Scaling factors -- supports only 1
    std::vector<float> %s_scales(%s_num_src, 1);

    // Create an array of memory descriptors and memory objects for src tensors.
    std::vector<memory::desc> %s_src_md;
    std::vector<memory> %s_src_mem;

"""%(layerNumber, layerNumber, node.outputs[0], num_src, node.outputs[0], 
    node.outputs[0], node.outputs[0], node.outputs[0]))
                for i in range(num_src):
                    f.write( \
"""    
    %s_src_md.push_back(%s_dst_md);
    %s_src_mem.push_back(%s_dst_memory);

"""%(node.outputs[0], node.inputs[i], node.outputs[0], node.inputs[i]))
                f.write( \
"""
    // Create primitive descriptor.
    auto %s_pd = sum::primitive_desc(%s_scales, %s_src_md, eng);

    // Create memory object for dst.
    auto %s_dst_memory = memory(%s_pd.dst_desc(), eng);

    // Adding sum primitive to net 
    net.push_back(sum(%s_pd));
    net_args.push_back({{ZENDNN_ARG_DST, %s_dst_memory}});
"""%(node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], 
    node.outputs[0], node.outputs[0], node.outputs[0]))
                for i in range(num_src):
                    f.write( \
"""
    net_args.push_back({{ZENDNN_ARG_MULTIPLE_SRC + %d, %s_src_mem[%d]}});

    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - Sum Setup Complete");
    // End: Model Layer %d - Sum
"""%(i, node.outputs[0], i, layerNumber, layerNumber))
                layerNumber += 1
            elif node.type == 'softmax':
                axis = node.attr.get('axis')
                f.write( \
"""
    // Start: Model Layer %d - softMax
    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - softmax Setup");

    // Softmax axis.
    const int sm_axis = %d;

    // Create softmax primitive
    auto %s_desc = softmax_forward::desc(prop_kind::forward_inference, %s_dst_memory.get_desc(), sm_axis);
    auto %s_pd = softmax_forward::primitive_desc(%s_desc, eng);

    // softmax - dst memory
    auto %s_dst_memory = memory(%s_pd.dst_desc(), eng);

    // Create a softmax primitive and add it to the net
    net.push_back(softmax_forward(%s_pd));
    net_args.push_back({{ZENDNN_ARG_SRC, %s_dst_memory},
                        {ZENDNN_ARG_DST, %s_dst_memory}});
    zendnnInfo(ZENDNN_TESTLOG, "Model Layer %d - softmax Setup Complete");
    // End: Model Layer %d- softMax
"""%(layerNumber, layerNumber, axis, node.outputs[0], node.inputs[0], node.outputs[0], 
    node.outputs[0], node.outputs[0], node.outputs[0], node.outputs[0], node.inputs[0], node.outputs[0], layerNumber, layerNumber))
                layerNumber += 1
        f.write( \
"""
    {
        unsigned int magic = 0;
        fread(&magic, 1, sizeof(magic), fp__variables);
        fclose(fp__variables);
        if (magic != 0xf00dd1e2)
        {
            printf("ERROR: invalid eoff magic in %s\\n", binaryFilename);
            return -1;
        }
    }

    // Execute primitives
    // For this example, the net is executed multiple times and each execution is timed individually.
    float AverageTime = 0;
    for (int j = 0; j < times; ++j)
    {
        auto start_exec_time = std::chrono::high_resolution_clock::now();
        assert(net.size() == net_args.size() && "something is missing");
        for (size_t i = 0; i < net.size(); ++i)
        {
            net.at(i).execute(s, net_args.at(i));
        }
        // Wait for the computation to finalize.
        s.wait();
        auto end_exec_time = std::chrono::high_resolution_clock::now();
        auto exec_duration = end_exec_time - start_exec_time;
        auto exec_in_millis = std::chrono::duration_cast<std::chrono::milliseconds>(exec_duration);
        printf("Execution:%d -- \t%.8f ms\\n", j, (float)exec_in_millis.count());

        // Read output from engine
        auto start_mem_time = std::chrono::high_resolution_clock::now();
""")
        for tensor in graph.outputs:
            f.write ( \
"""
        read_from_zendnn_memory(user_dst.data(), %s_dst_memory);
"""%(tensor.name))
        f.write( \
"""
        auto end_mem_time = std::chrono::high_resolution_clock::now();
        auto exec_mem_duration = end_mem_time - start_mem_time;
        auto mem_in_millis = std::chrono::duration_cast<std::chrono::milliseconds>(exec_mem_duration);
        //printf("Mem Transfer:%d -- \t%.8f ms\\n", j, (float)mem_in_millis.count());

        AverageTime += (float)exec_in_millis.count() + (float)mem_in_millis.count();
    }
    printf("\\nAvg Inference Time -- \t%.8f ms\\n",(float)(AverageTime/times));
        return 0;
}

int main(int argc, const char **argv)
{
    zendnnInfo(ZENDNN_TESTLOG, "zendnn_app test starts");

    // check command-line usage
    if (argc != 3)
    {
        printf(
            "\\n"
            "Usage: ZENDNN_LOG_OPTS=ALL:5 ZENDNN_VERBOSE=1 ./zendnn_app [weights.bin] [imageName]\\n"
            "\\n"
            "   <weights.bin>: weights file to be used for the inference\\n"
            "   <imageName>: image file to be used for the inference\\n"
            "\\n");
        return -1;
    }

    engine::kind engine_kind = zendnn::engine::kind::cpu;
    const char *weights = argv[1];
    const char *image = argv[2];
    int NumExecution = 10;

    try
    {
        auto begin = chrono::duration_cast<chrono::milliseconds>(
                         chrono::steady_clock::now().time_since_epoch())
                         .count();

        int app_status = model_setup(engine_kind, weights, image, NumExecution);

        auto end = chrono::duration_cast<chrono::milliseconds>(
                       chrono::steady_clock::now().time_since_epoch())
                       .count();
        zendnnInfo(ZENDNN_TESTLOG, "Use time ", (end - begin) / (NumExecution + 0.0));
    }

    catch (error &e)
    {
        std::cerr << "status: " << e.status << std::endl;
        std::cerr << "message: " << e.message << std::endl;
    }
    zendnnInfo(ZENDNN_TESTLOG, "zendnn_app test ends\\n");
    return 0;
}
""")

def generateBinary(graph,fileName):
    VARIABLES_FILE_MAGIC = 0xF00DD1E0
    VARIABLES_DATA_MAGIC = 0xF00DD1E1
    VARIABLES_EOFF_MAGIC = 0xF00DD1E2
    print('creating ' + fileName + ' ...')
    with open(fileName, 'wb') as f:
        f.write(struct.pack('I', VARIABLES_FILE_MAGIC))
        for tensor in graph.initializers:
            binary = graph.binaries[tensor.name]
            f.write(struct.pack('II', VARIABLES_DATA_MAGIC, len(binary)))
            f.write(binary)
        f.write(struct.pack('I', VARIABLES_EOFF_MAGIC))

def generateCode(graph,argmaxOutput,outputFolder):
    if not os.path.isdir(outputFolder):
        os.mkdir(outputFolder)
        os.mkdir(outputFolder + '/data')
        os.mkdir(outputFolder + '/source')
    generateCMakeFiles(graph,outputFolder)
    generateBinary(graph,outputFolder + '/data/weights.bin')
    generateDeployCPP(graph,outputFolder + '/source/zentest.cpp')

def main():
    usage = """
Usage: python nnir_to_zendnn.py <nnirInputFolder> <outputFolder>

"""
    pos = 1
    argmaxOutput = None
    while len(sys.argv[pos:]) >= 2 and sys.argv[pos][:2] == '--':
        if sys.argv[pos] != '--help':
            print('ERROR: invalid option: %s' % (sys.argv[pos]))
        print(usage)
        sys.exit(1)
        pos = pos + 2
    if len(sys.argv[pos:]) < 2:
        print(usage)
        sys.exit(1)
    inputFolder = sys.argv[pos]
    outputFolder = sys.argv[pos+1]
    print('reading IR model from ' + inputFolder + ' ...')
    graph = ir.IrGraph(True)
    graph.fromFile(inputFolder)
    print('creating C code in ' + outputFolder + ' ...')
    generateCode(graph,argmaxOutput,outputFolder)


if __name__ == '__main__':
    main()



__author__      = "Kiriti Nagesh Gowda"
__copyright__   = "Copyright 2019 - 2024, AMD MIVisionX"
__credits__     = ["Mike Schmit; Hansel Yang; Lakshmi Kumar;"]
__license__     = "MIT"
__version__     = "1.1"
__maintainer__  = "Kiriti Nagesh Gowda"
__email__       = "mivisionx.support@amd.com"
__status__      = "Shipping"
__script_name__ = "MIVisionX Inference Analyzer"

import argparse
import os
import sys
import ctypes
import cv2
import time
import numpy
import statistics
import numpy as np
from numpy.ctypeslib import ndpointer
from inference_control import inference_control
from PyQt5 import QtWidgets

# global variables
FP16inference = False
verbosePrint = False
labelNames = None
colors =[
        (0,153,0),        # Top1
        (153,153,0),      # Top2
        (153,76,0),       # Top3
        (0,128,255),      # Top4
        (255,102,102),    # Top5
        ]
# error check calls
def ERROR_CHECK(call):
    status = call
    if(status != 0):
        exit(status)

# AMD Neural Net python wrapper
class AnnAPI:
    def __init__(self,library):
        self.lib = ctypes.cdll.LoadLibrary(library)
        self.annQueryInference = self.lib.annQueryInference
        self.annQueryInference.restype = ctypes.c_char_p
        self.annQueryInference.argtypes = []
        self.annCreateInference = self.lib.annCreateInference
        self.annCreateInference.restype = ctypes.c_void_p
        self.annCreateInference.argtypes = [ctypes.c_char_p]
        self.annReleaseInference = self.lib.annReleaseInference
        self.annReleaseInference.restype = ctypes.c_int
        self.annReleaseInference.argtypes = [ctypes.c_void_p]
        self.annCopyToInferenceInput = self.lib.annCopyToInferenceInput
        self.annCopyToInferenceInput.restype = ctypes.c_int
        self.annCopyToInferenceInput.argtypes = [ctypes.c_void_p, ndpointer(ctypes.c_float, flags="C_CONTIGUOUS"), ctypes.c_size_t, ctypes.c_bool]
        self.annCopyFromInferenceOutput = self.lib.annCopyFromInferenceOutput
        self.annCopyFromInferenceOutput.restype = ctypes.c_int
        self.annCopyFromInferenceOutput.argtypes = [ctypes.c_void_p, ndpointer(ctypes.c_float, flags="C_CONTIGUOUS"), ctypes.c_size_t]
        self.annRunInference = self.lib.annRunInference
        self.annRunInference.restype = ctypes.c_int
        self.annRunInference.argtypes = [ctypes.c_void_p, ctypes.c_int]
        print('OK: AMD VX NN API found "' + self.annQueryInference().decode("utf-8") + '" as configuration in ' + library)

# classifier definition
class annieObjectWrapper():
    def __init__(self, annpythonlib, weightsfile):
        self.api = AnnAPI(annpythonlib)
        input_info,output_info,empty = self.api.annQueryInference().decode("utf-8").split(';')
        input,name,n_i,c_i,h_i,w_i = input_info.split(',')
        outputCount = output_info.split(",")
        stringcount = len(outputCount)
        if stringcount == 6:
            output,opName,n_o,c_o,h_o,w_o = output_info.split(',')
        else:
            output,opName,n_o,c_o= output_info.split(',')
            h_o = '1'; w_o  = '1'
        self.hdl = self.api.annCreateInference(weightsfile.encode('utf-8'))
        self.dim = (int(w_i),int(h_i))
        self.outputDim = (int(n_o),int(c_o),int(h_o),int(w_o))

    def __del__(self):
        self.api.annReleaseInference(self.hdl)

    def runInference(self, img, out):
        # create input.f32 file
        img_r = img[:,:,0]
        img_g = img[:,:,1]
        img_b = img[:,:,2]
        img_t = np.concatenate((img_r, img_g, img_b), 0)    
        # copy input f32 to inference input
        status = self.api.annCopyToInferenceInput(self.hdl, np.ascontiguousarray(img_t, dtype=np.float32), (img.shape[0]*img.shape[1]*3*4), 0)
        if(status):
                print('ERROR: AMD VX NN CopyToInferenceInput Failed ')
        # run inference
        status = self.api.annRunInference(self.hdl, 1)
        if(status):
                print('ERROR: AMD VX NN RunInference Failed ')
        # copy output f32
        status = self.api.annCopyFromInferenceOutput(self.hdl, np.ascontiguousarray(out, dtype=np.float32), out.nbytes)
        if(status):
                print('ERROR: AMD VX NN CopyFromInferenceOutput Failed ')
        return out

    def classify(self, img):
        # create output.f32 buffer
        out_buf = bytearray(self.outputDim[0]*self.outputDim[1]*self.outputDim[2]*self.outputDim[3]*4)
        out = np.frombuffer(out_buf, dtype=numpy.float32)
        # run inference & receive output
        output = self.runInference(img, out)
        return output

# process classification output function
def processClassificationOutput(inputImage, modelName, modelOutput):
    # post process output file
    start = time.time()
    softmaxOutput = np.float32(modelOutput)
    topIndex = []
    topLabels = []
    topProb = []
    for x in softmaxOutput.argsort()[-5:]:
        topIndex.append(x)
        topLabels.append(labelNames[x])
        topProb.append(softmaxOutput[x])
    end = time.time()
    if(verbosePrint):
        print('%30s' % 'Processed results in ', str((end - start)*1000), 'ms')

    # display output
    start = time.time()
    # initialize the result image
    resultImage = np.zeros((250, 525, 3), dtype="uint8")
    resultImage.fill(255)
    cv2.putText(resultImage, 'MIVisionX Object Classification', (25,  25),cv2.FONT_HERSHEY_SIMPLEX, 1.0, (0, 0, 0), 2)
    topK = 1   
    for i in reversed(range(5)):
        txt =  topLabels[i]
        conf = topProb[i]
        txt = 'Top'+str(topK)+':'+txt+' '+str(int(round((conf*100), 0)))+'%' 
        size = cv2.getTextSize(txt, cv2.FONT_HERSHEY_SIMPLEX, 0.5, 1)
        t_height = size[0][1]
        textColor = (colors[topK - 1])
        cv2.putText(resultImage,txt,(45,t_height+(topK*30+40)),cv2.FONT_HERSHEY_SIMPLEX,0.5,textColor,1)
        topK = topK + 1
    end = time.time()
    if(verbosePrint):
        print('%30s' % 'Processed results image in ', str((end - start)*1000), 'ms')

    return resultImage, topIndex, topProb

# MIVisionX Classifier
if __name__ == '__main__':
    
    if len(sys.argv) == 1:
        app = QtWidgets.QApplication(sys.argv)
        panel = inference_control()
        app.exec_()
        modelFormat = (str)(panel.model_format)
        modelName = (str)(panel.model_name)
        modelLocation = (str)(panel.model)
        modelInputDims = (str)(panel.input_dims)
        modelOutputDims = (str)(panel.output_dims)
        label = (str)(panel.label)
        outputDir = (str)(panel.output)
        imageDir = (str)(panel.image)
        imageVal = (str)(panel.val)
        hierarchy = (str)(panel.hier)
        inputAdd = (str)(panel.add)
        inputMultiply = (str)(panel.multiply)
        fp16 = (str)(panel.fp16)
        replaceModel = (str)(panel.replace)
        verbose = (str)(panel.verbose)
    else:
        parser = argparse.ArgumentParser()
        parser.add_argument('--model_format',       type=str, required=True,    help='pre-trained model format, options:caffe/onnx/nnef [required]')
        parser.add_argument('--model_name',         type=str, required=True,    help='model name                             [required]')
        parser.add_argument('--model',              type=str, required=True,    help='pre_trained model file/folder          [required]')
        parser.add_argument('--model_input_dims',   type=str, required=True,    help='c,h,w - channel,height,width           [required]')
        parser.add_argument('--model_output_dims',  type=str, required=True,    help='c,h,w - channel,height,width           [required]')
        parser.add_argument('--label',              type=str, required=True,    help='labels text file                       [required]')
        parser.add_argument('--output_dir',         type=str, required=True,    help='output dir to store ADAT results       [required]')
        parser.add_argument('--image_dir',          type=str, required=True,    help='image directory for analysis           [required]')
        parser.add_argument('--image_val',          type=str, default='',       help='image list with ground truth           [optional]')
        parser.add_argument('--hierarchy',          type=str, default='',       help='AMD proprietary hierarchical file      [optional]')
        parser.add_argument('--add',                type=str, default='',       help='input preprocessing factor [optional - default:[0,0,0]]')
        parser.add_argument('--multiply',           type=str, default='',       help='input preprocessing factor [optional - default:[1,1,1]]')
        parser.add_argument('--fp16',               type=str, default='no',     help='quantize to FP16          [optional - default:no]')
        parser.add_argument('--replace',            type=str, default='no',     help='replace/overwrite model   [optional - default:no]')
        parser.add_argument('--verbose',            type=str, default='no',     help='verbose                   [optional - default:no]')
        args = parser.parse_args()
        
        # get arguments
        modelFormat = args.model_format
        modelName = args.model_name
        modelLocation = args.model
        modelInputDims = args.model_input_dims
        modelOutputDims = args.model_output_dims
        label = args.label
        outputDir = args.output_dir
        imageDir = args.image_dir
        imageVal = args.image_val
        hierarchy = args.hierarchy
        inputAdd = args.add
        inputMultiply = args.multiply
        fp16 = args.fp16
        replaceModel = args.replace
        verbose = args.verbose
    # set verbose print
    if(verbose != 'no'):
        verbosePrint = True

    # set fp16 inference turned on/off
    if(fp16 != 'no'):
        FP16inference = True

    # set paths
    modelCompilerPath = '/opt/rocm/libexec/mivisionx/model_compiler/python'
    ADATPath= '/opt/rocm/libexec/mivisionx/toolkit/amd_data_analysis_toolkit/classification'
    setupDir = '~/.mivisionx-inference-analyzer'
    analyzerDir = os.path.expanduser(setupDir)
    modelDir = analyzerDir+'/'+modelName+'_dir'
    nnirDir = modelDir+'/nnir-files'
    openvxDir = modelDir+'/openvx-files'
    modelBuildDir = modelDir+'/build'
    adatOutputDir = os.path.expanduser(outputDir)
    inputImageDir = os.path.expanduser(imageDir)
    trainedModel = os.path.expanduser(modelLocation)
    labelText = os.path.expanduser(label)
    hierarchyText = os.path.expanduser(hierarchy)
    imageValText = os.path.expanduser(imageVal)
    pythonLib = modelBuildDir+'/libannpython.so'
    weightsFile = openvxDir+'/weights.bin'
    finalImageResultsFile = modelDir+'/imageResultsFile.csv'

    # get input & output dims
    str_c_i, str_h_i, str_w_i = modelInputDims.split(',')
    c_i = int(str_c_i); h_i = int(str_h_i); w_i = int(str_w_i)
    str_c_o, str_h_o, str_w_o = modelOutputDims.split(',')
    c_o = int(str_c_o); h_o = int(str_h_o); w_o = int(str_w_o)

    # input pre-processing values
    Ax=[0,0,0]
    if(inputAdd != ''):
        Ax = [float(item) for item in inputAdd.strip("[]").split(',')]
    Mx=[1,1,1]
    if(inputMultiply != ''):
        Mx = [float(item) for item in inputMultiply.strip("[]").split(',')]

    # check pre-trained model
    if(not os.path.isfile(trainedModel) and modelFormat != 'nnef' ):
        print("\nPre-Trained Model not found, check argument --model\n")
        quit()

    # check for label file
    if (not os.path.isfile(labelText)):
        print("\nlabels.txt not found, check argument --label\n")
        quit()
    else:
        fp = open(labelText, 'r')
        labelNames = fp.readlines()
        labelNames = [line.rstrip('\n') for line in labelNames]
        fp.close()

    # MIVisionX setup
    if(os.path.exists(analyzerDir)):
        print("\nMIVisionX Inference Analyzer\n")
        # replace old model or throw error
        if(replaceModel == 'yes'):
            ERROR_CHECK(os.system('rm -rf '+modelDir))
        elif(os.path.exists(modelDir)):
            print("OK: Model exists")

    else:
        print("\nMIVisionX Inference Analyzer Created\n")
        ERROR_CHECK(os.system('(cd ; mkdir .mivisionx-inference-analyzer)'))

    # Setup Text File for Demo
    if (not os.path.isfile(analyzerDir + "/setupFile.txt")):
        f = open(analyzerDir + "/setupFile.txt", "w")
        f.write(modelFormat + ';' + modelName + ';' + modelLocation + ';' + modelInputDims + ';' + modelOutputDims + ';' + label + ';' + outputDir + ';' + imageDir + ';' + imageVal + ';' + hierarchy + ';' + str(Ax).strip('[]').replace(" ","") + ';' + str(Mx).strip('[]').replace(" ","") + ';' + fp16 + ';' + replaceModel + ';' + verbose)
        f.close()
    else:
        count = len(open(analyzerDir + "/setupFile.txt").readlines())
        if count < 10:
            with open(analyzerDir + "/setupFile.txt", "r") as fin:
                data = fin.read().splitlines(True)
                modelList = []
                for i in range(len(data)):
                    modelList.append(data[i].split(';')[1])
                if modelName not in modelList:
                    f = open(analyzerDir + "/setupFile.txt", "a")
                    f.write("\n" + modelFormat + ';' + modelName + ';' + modelLocation + ';' + modelInputDims + ';' + modelOutputDims + ';' + label + ';' + outputDir + ';' + imageDir + ';' + imageVal + ';' + hierarchy + ';' + str(Ax).strip('[]').replace(" ","") + ';' + str(Mx).strip('[]').replace(" ","") + ';' + fp16 + ';' + replaceModel + ';' + verbose)
                    f.close()
        else:
            with open(analyzerDir + "/setupFile.txt", "r") as fin:
                data = fin.read().splitlines(True)
            delModelName = data[0].split(';')[1]
            delmodelPath = analyzerDir + '/' + delModelName + '_dir'
            if(os.path.exists(delmodelPath)): 
                ERROR_CHECK(os.system('rm -rf ' + delmodelPath))
            with open(analyzerDir + "/setupFile.txt", "w") as fout:
                fout.writelines(data[1:])
            with open(analyzerDir + "/setupFile.txt", "a") as fappend:
                fappend.write("\n" + modelFormat + ';' + modelName + ';' + modelLocation + ';' + modelInputDims + ';' + modelOutputDims + ';' + label + ';' + outputDir + ';' + imageDir + ';' + imageVal + ';' + hierarchy + ';' + str(Ax).strip('[]').replace(" ","") + ';' + str(Mx).strip('[]').replace(" ","") + ';' + fp16 + ';' + replaceModel + ';' + verbose)
                fappend.close()

    # Compile Model and generate python .so files
    if (replaceModel == 'yes' or not os.path.exists(modelDir)):
        ERROR_CHECK(os.system('mkdir '+modelDir))
        if(os.path.exists(modelDir)):
            # convert to NNIR
            if(modelFormat == 'caffe'):
                ERROR_CHECK(os.system('(cd '+modelDir+'; python3 '+modelCompilerPath+'/caffe_to_nnir.py '+trainedModel+' nnir-files --input-dims 1,'+modelInputDims+' )'))
            elif(modelFormat == 'onnx'):
                ERROR_CHECK(os.system('(cd '+modelDir+'; python3 '+modelCompilerPath+'/onnx_to_nnir.py '+trainedModel+' nnir-files --input-dims 1,'+modelInputDims+' )'))
            elif(modelFormat == 'nnef'):
                ERROR_CHECK(os.system('(cd '+modelDir+'; python3 '+modelCompilerPath+'/nnef_to_nnir.py '+trainedModel+' nnir-files )'))
            else:
                print("ERROR: Neural Network Format Not supported, use caffe/onnx/nnef in arugment --model_format")
                quit()
            # convert the model to FP16
            if(FP16inference):
                ERROR_CHECK(os.system('(cd '+modelDir+'; python3 '+modelCompilerPath+'/nnir_update.py --convert-fp16 1 --fuse-ops 1 nnir-files nnir-files)'))
                print("\nModel Quantized to FP16\n")
            # convert to openvx
            if(os.path.exists(nnirDir)):
                ERROR_CHECK(os.system('(cd '+modelDir+'; python3 '+modelCompilerPath+'/nnir_to_openvx.py nnir-files openvx-files)'))
            else:
                print("ERROR: Converting Pre-Trained model to NNIR Failed")
                quit()
            
            # build model
            if(os.path.exists(openvxDir)):
                ERROR_CHECK(os.system('mkdir '+modelBuildDir))
            else:
                print("ERROR: Converting NNIR to OpenVX Failed")
                quit()
    ERROR_CHECK(os.system('(cd '+modelBuildDir+'; cmake ../openvx-files; make; ./anntest ../openvx-files/weights.bin )'))
    print("\nSUCCESS: Converting Pre-Trained model to MIVisionX Runtime successful\n")
    
    #else:
        #print("ERROR: MIVisionX Inference Analyzer Failed")
        #quit()

    #logos
    AMD_Epyc_Black_resize = cv2.imread("../../samples/images/amd-epyc-black-resize.png")
    AMD_ROCm_Black_resize = cv2.imread("../../samples/images/rocm-black-resize.png")

    # opencv display window
    windowInput = "MIVisionX Inference Analyzer - Input Image"
    windowResult = "MIVisionX Inference Analyzer - Results"
    windowProgress = "MIVisionX Inference Analyzer - Progress"
    cv2.namedWindow(windowInput, cv2.WINDOW_GUI_EXPANDED)
    cv2.resizeWindow(windowInput, 800, 800)

    # create inference classifier
    classifier = annieObjectWrapper(pythonLib, weightsFile)

    # check for image val text
    totalImages = 0
    if(imageVal == ''):
        print("\nFlow without Image Validation Text..Creating a file with no ground truths\n")
        imageList = os.listdir(inputImageDir)
        imageList.sort()
        imageValText = os.getcwd() + '/imageValTxt.txt'
        fp = open(imageValText , 'w')
        for imageFile in imageList:
            fp.write(imageFile + " -1" + "\n")

    
    if (not os.path.isfile(imageValText)):
        print("\nImage Validation Text not found, check argument --image_val\n")
        quit()
    else:
        fp = open(imageValText, 'r')
        imageValidation = fp.readlines()
        fp.close()
        totalImages = len(imageValidation)

    # original std out location 
    orig_stdout = sys.stdout
    # setup results output file
    sys.stdout = open(finalImageResultsFile,'w')
    print('Image File Name,Ground Truth Label,Output Label 1,Output Label 2,Output Label 3,\
            Output Label 4,Output Label 5,Prob 1,Prob 2,Prob 3,Prob 4,Prob 5')
    sys.stdout = orig_stdout

    # process images
    correctTop5 = 0; correctTop1 = 0; wrong = 0; noGroundTruth = 0
    totalTimeForInference = 0
    totalCorners = 0
    listOfCorners = []
    for x in range(totalImages):
        imageFileName,grountTruth = imageValidation[x].split(' ')
        groundTruthIndex = int(grountTruth)
        imageFile = os.path.expanduser(inputImageDir+'/'+imageFileName)
        if (not os.path.isfile(imageFile)):
            print('Image File - '+imageFile+' not found')
            quit()
        else:
            # read image
            start = time.time()
            frame = cv2.imread(imageFile)
            end = time.time()
            totalTimeForInference += ((end - start)*1000)
            if(verbosePrint):
                print('%30s' % 'Read Image in ', str((end - start)*1000), 'ms')
                
            # image complexity detection
            start = time.time()
            gray = cv2.cvtColor(frame,cv2.COLOR_BGR2GRAY)
            gray = np.float32(gray)
            dst = cv2.cornerHarris(gray,2,3,0.04)
            numCorners = np.sum(dst > 0.01 * dst.max())
            listOfCorners.append(numCorners)
            totalCorners += numCorners
            end = time.time()
            if(verbosePrint):
                print('%30s' % 'Image complexity detection in ', str((end - start)*1000), 'ms')

            # resize and process frame
            start = time.time()
            if(numCorners <= 3500):
                resizedFrame = cv2.resize(frame, (w_i,h_i), interpolation = cv2.INTER_LINEAR)
            else:
                resizedFrame = cv2.resize(frame, (w_i,h_i), interpolation = cv2.INTER_AREA)
            RGBframe = cv2.cvtColor(resizedFrame, cv2.COLOR_BGR2RGB)
            if(inputAdd != '' or inputMultiply != ''):
                pFrame = np.zeros(RGBframe.shape).astype('float32')
                for i in range(RGBframe.shape[2]):
                    pFrame[:,:,i] = RGBframe.copy()[:,:,i] * Mx[i] + Ax[i]
                RGBframe = pFrame
            end = time.time()
            totalTimeForInference += ((end - start)*1000)
            if(verbosePrint):
                print('%30s' % 'Input pre-processed in ', str((end - start)*1000), 'ms')

            # run inference
            start = time.time()
            output = classifier.classify(RGBframe)
            end = time.time()
            totalTimeForInference += ((end - start)*1000)
            if(verbosePrint):
                print('%30s' % 'Executed Model in ', str((end - start)*1000), 'ms')

            # process output and display
            resultImage, topIndex, topProb = processClassificationOutput(resizedFrame, modelName, output)
            # copy logos to output window
            frameWindow = cv2.resize(frame, (800,800))
            frameWindow[0:AMD_ROCm_Black_resize.shape[0], 0:AMD_ROCm_Black_resize.shape[1]] = AMD_ROCm_Black_resize
            frameWindow[0:AMD_Epyc_Black_resize.shape[0], 607: 607 + AMD_Epyc_Black_resize.shape[1]] = AMD_Epyc_Black_resize
            start = time.time()
            cv2.imshow(windowInput, frameWindow)
            cv2.imshow(windowResult, resultImage)
            end = time.time()
            if(verbosePrint):
                print('%30s' % 'Processed display in ', str((end - start)*1000), 'ms')

            # write image results to a file
            start = time.time()
            sys.stdout = open(finalImageResultsFile,'a')
            print(imageFileName+','+str(groundTruthIndex)+','+str(topIndex[4])+
                ','+str(topIndex[3])+','+str(topIndex[2])+','+str(topIndex[1])+','+str(topIndex[0])+','+str(topProb[4])+
                ','+str(topProb[3])+','+str(topProb[2])+','+str(topProb[1])+','+str(topProb[0]))
            sys.stdout = orig_stdout
            end = time.time()
            if(verbosePrint):
                print('%30s' % 'Image result saved in ', str((end - start)*1000), 'ms')

            # create progress image
            start = time.time()
            progressImage = np.zeros((400, 500, 3), dtype="uint8")
            progressImage.fill(255)
            cv2.putText(progressImage, 'Inference Analyzer Progress', (25,  25),cv2.FONT_HERSHEY_SIMPLEX, 1.0, (0, 0, 0), 2)
            size = cv2.getTextSize(modelName, cv2.FONT_HERSHEY_SIMPLEX, 0.7, 2)
            t_width = size[0][0]
            t_height = size[0][1]
            headerX_start = int(250 -(t_width/2))
            cv2.putText(progressImage,modelName,(headerX_start,t_height+(20+40)),cv2.FONT_HERSHEY_SIMPLEX,0.7,(0,0,0),2)
            txt = 'Processed: '+str(x+1)+' of '+str(totalImages)
            size = cv2.getTextSize(txt, cv2.FONT_HERSHEY_SIMPLEX, 0.5, 2)
            cv2.putText(progressImage,txt,(50,t_height+(60+40)),cv2.FONT_HERSHEY_SIMPLEX,0.5,(0,0,0),1)
            # progress bar
            cv2.rectangle(progressImage, (50,150), (450,180), (192,192,192), -1)
            progressWidth = int(50+ ((400*(x+1))/totalImages))
            cv2.rectangle(progressImage, (50,150), (progressWidth,180), (255,204,153), -1)
            percentage = int(((x+1)/float(totalImages))*100)
            pTxt = 'progress: '+str(percentage)+'%'
            cv2.putText(progressImage,pTxt,(175,170),cv2.FONT_HERSHEY_SIMPLEX,0.5,(0,0,0),1)

            if(groundTruthIndex == topIndex[4]):
                correctTop1 = correctTop1 + 1
                correctTop5 = correctTop5 + 1
            elif(groundTruthIndex == topIndex[3] or groundTruthIndex == topIndex[2] or groundTruthIndex == topIndex[1] or groundTruthIndex == topIndex[0]):
                correctTop5 = correctTop5 + 1
            elif(groundTruthIndex == -1):
                noGroundTruth = noGroundTruth + 1
            else:
                wrong = wrong + 1

            # top 1 progress
            cv2.rectangle(progressImage, (50,200), (450,230), (192,192,192), -1)
            progressWidth = int(50 + ((400*correctTop1)/totalImages))
            cv2.rectangle(progressImage, (50,200), (progressWidth,230), (0,153,0), -1)
            percentage = int((correctTop1/float(totalImages))*100)
            pTxt = 'Top1: '+str(percentage)+'%'
            cv2.putText(progressImage,pTxt,(195,220),cv2.FONT_HERSHEY_SIMPLEX,0.5,(0,0,0),1)
            # top 5 progress
            cv2.rectangle(progressImage, (50,250), (450,280), (192,192,192), -1)
            progressWidth = int(50+ ((400*correctTop5)/totalImages))
            cv2.rectangle(progressImage, (50,250), (progressWidth,280), (0,255,0), -1)
            percentage = int((correctTop5/float(totalImages))*100)
            pTxt = 'Top5: '+str(percentage)+'%'
            cv2.putText(progressImage,pTxt,(195,270),cv2.FONT_HERSHEY_SIMPLEX,0.5,(0,0,0),1)
            # wrong progress
            cv2.rectangle(progressImage, (50,300), (450,330), (192,192,192), -1)
            progressWidth = int(50+ ((400*wrong)/totalImages))
            cv2.rectangle(progressImage, (50,300), (progressWidth,330), (0,0,255), -1)
            percentage = int((wrong/float(totalImages))*100)
            pTxt = 'Mismatch: '+str(percentage)+'%'
            cv2.putText(progressImage,pTxt,(175,320),cv2.FONT_HERSHEY_SIMPLEX,0.5,(0,0,0),1)
            # no ground truth progress
            cv2.rectangle(progressImage, (50,350), (450,380), (192,192,192), -1)
            progressWidth = int(50+ ((400*noGroundTruth)/totalImages))
            cv2.rectangle(progressImage, (50,350), (progressWidth,380), (0,255,255), -1)
            percentage = int((noGroundTruth/float(totalImages))*100)
            pTxt = 'Ground Truth unavailable: '+str(percentage)+'%'
            cv2.putText(progressImage,pTxt,(125,370),cv2.FONT_HERSHEY_SIMPLEX,0.5,(0,0,0),1)
            
            cv2.imshow(windowProgress, progressImage)
            end = time.time()
            if(verbosePrint):
                print('%30s' % 'Progress image created in ', str((end - start)*1000), 'ms')
                print('%30s' % 'Images Processed: ', str(x), '\n\n')

            # exit on ESC
            key = cv2.waitKey(2)
            if key == 27: 
                break

    # Inference Analyzer Successful
    print("\nSUCCESS: Images Inferenced with the Model\n")
    timePerInference = float (totalTimeForInference / totalImages)
    print('%30s' % 'Time per image Inference ', str(timePerInference), 'ms')
    avgCornersPerImage = float(totalCorners / totalImages )
    print('%30s' % 'Avg number of corners per image: ', str(avgCornersPerImage))
    print('%30s' % 'Mean number of corners per image: ', str(statistics.mean(listOfCorners)))
    cv2.destroyWindow(windowInput)
    cv2.destroyWindow(windowResult)

    # Create ADAT folder and file
    print("\nADAT tool called to create the analysis toolkit\n")
    if(not os.path.exists(adatOutputDir)):
        ERROR_CHECK(os.system('mkdir -p ' + adatOutputDir))
    
    if(hierarchy == ''):
        ERROR_CHECK(os.system('python '+ADATPath+'/generate-visualization.py --inference_results '+finalImageResultsFile+
        ' --image_dir '+inputImageDir+' --label '+labelText+' --model_name '+modelName+' --output_dir '+adatOutputDir+' --output_name '+modelName+'-ADAT'))
    else:
        ERROR_CHECK(os.system('python '+ADATPath+'/generate-visualization.py --inference_results '+finalImageResultsFile+
        ' --image_dir '+inputImageDir+' --label '+labelText+' --hierarchy '+hierarchyText+' --model_name '+modelName+' --output_dir '+adatOutputDir+' --output_name '+modelName+'-ADAT'))
    print("\nSUCCESS: Image Analysis Toolkit Created\n")
    print("Press ESC to exit or close progress window\n")

    # Wait to quit
    while True:
        key = cv2.waitKey(2)
        if key == 27:
            cv2.destroyAllWindows()
            break        
        if cv2.getWindowProperty(windowProgress,cv2.WND_PROP_VISIBLE) < 1:        
            break

    outputHTMLFile = os.path.expanduser(adatOutputDir+'/'+modelName+'-ADAT-toolKit/index.html')
    ERROR_CHECK(os.system('firefox '+outputHTMLFile))



import os
from PyQt5 import QtWidgets, uic

class inference_control(QtWidgets.QMainWindow):
    def __init__(self, parent=None):
        super(inference_control, self).__init__(parent)
        self.model_format = ''
        self.model_name = ''
        self.model = ''
        self.input_dims = ''
        self.output_dims = ''
        self.label = ''
        self.output = ''
        self.image = ''
        self.val = ''
        self.hier = ''
        self.add = '0,0,0'
        self.multiply = '1,1,1'
        self.fp16 = 'no'
        self.replace = 'no'
        self.verbose = 'no'
        self.runningState = False
        self.initUI()

    def initUI(self):
        uic.loadUi("inference_control.ui", self)
        #self.setStyleSheet("background-color: white")
        self.upload_comboBox.activated.connect(self.fromFile)
        self.file_pushButton.clicked.connect(self.browseFile)
        self.output_pushButton.clicked.connect(self.browseOutput)
        self.label_pushButton.clicked.connect(self.browseLabel)
        self.image_pushButton.clicked.connect(self.browseImage)
        self.val_pushButton.clicked.connect(self.browseVal)
        self.hier_pushButton.clicked.connect(self.browseHier)
        self.run_pushButton.clicked.connect(self.runConfig)
        self.file_lineEdit.textChanged.connect(self.checkInput)
        self.name_lineEdit.textChanged.connect(self.checkInput)
        self.idims_lineEdit.textChanged.connect(self.checkInput)
        self.odims_lineEdit.textChanged.connect(self.checkInput)
        self.output_lineEdit.textChanged.connect(self.checkInput)
        self.label_lineEdit.textChanged.connect(self.checkInput)
        self.image_lineEdit.textChanged.connect(self.checkInput)
        self.image_lineEdit.textChanged.connect(self.checkInput)
        self.close_pushButton.clicked.connect(self.closeEvent)
        self.file_lineEdit.setPlaceholderText("File Directory for NNEF / Model File for CAFFE/ONNX [required]")
        self.name_lineEdit.setPlaceholderText("Model Name [required]")
        self.idims_lineEdit.setPlaceholderText("c,h,w [required]")
        self.odims_lineEdit.setPlaceholderText("c,h,w [required]")
        self.padd_lineEdit.setPlaceholderText("r,g,b [optional]")
        self.pmul_lineEdit.setPlaceholderText("r,g,b [optional]")
        self.output_lineEdit.setPlaceholderText("Output Directory [required]")
        self.label_lineEdit.setPlaceholderText("Label File [required]")
        self.image_lineEdit.setPlaceholderText("Image Folder [required]")
        self.val_lineEdit.setPlaceholderText("[optional]")
        self.hier_lineEdit.setPlaceholderText("[optional]")
        self.close_pushButton.setStyleSheet("color: white; background-color: darkRed")
        self.readSetupFile()
        self.show()

    def browseFile(self):
        if self.format_comboBox.currentText() == 'nnef':
            self.file_lineEdit.setText(QtWidgets.QFileDialog.getExistingDirectory(self, 'Open Folder', './'))
        else:
            filename, _filter = QtWidgets.QFileDialog.getOpenFileName(self, 'Open File', './')
            self.file_lineEdit.setText(filename)

    def browseOutput(self):
        self.output_lineEdit.setText(QtWidgets.QFileDialog.getExistingDirectory(self, 'Open Folder', './'))

    def browseLabel(self):
        filename, _filter = QtWidgets.QFileDialog.getOpenFileName(self, 'Open File', './', filter="Text files (*.txt)")
        self.label_lineEdit.setText(filename)

    def browseImage(self):
        self.image_lineEdit.setText(QtWidgets.QFileDialog.getExistingDirectory(self, 'Open Folder', './'))

    def browseVal(self):
        filename, _filter = QtWidgets.QFileDialog.getOpenFileName(self, 'Open File', './', filter="Text files (*.txt)")
        self.val_lineEdit.setText(filename)

    def browseHier(self):
        filename, _filter = QtWidgets.QFileDialog.getOpenFileName(self, 'Open File', './', '*.csv')
        self.hier_lineEdit.setText(filename)

    def readSetupFile(self):
        setupDir = '~/.mivisionx-inference-analyzer'
        analyzerDir = os.path.expanduser(setupDir)
        if os.path.isfile(analyzerDir + "/setupFile.txt"):
            for line in open(analyzerDir + "/setupFile.txt", "r"):
                token = line.split(';')
                if len(token) > 1:
                    modelName = token[1]
                    self.upload_comboBox.addItem(modelName)
            
    def fromFile(self):
        if self.upload_comboBox.currentIndex() == 0:
            self.name_lineEdit.setEnabled(True)
            self.file_lineEdit.setEnabled(True)
            self.idims_lineEdit.setEnabled(True)
            self.odims_lineEdit.setEnabled(True)
            self.label_lineEdit.setEnabled(True)
            self.output_lineEdit.setEnabled(True)
            self.image_lineEdit.setEnabled(True)
            self.val_lineEdit.setEnabled(True)
            self.hier_lineEdit.setEnabled(True)
            self.padd_lineEdit.setEnabled(True)
            self.pmul_lineEdit.setEnabled(True)
            self.fp16_checkBox.setEnabled(True)
            self.replace_checkBox.setEnabled(True)
            self.verbose_checkBox.setEnabled(True)
            self.file_pushButton.setEnabled(True)
            self.format_comboBox.setEnabled(True)
            self.output_pushButton.setEnabled(True)
            self.label_pushButton.setEnabled(True)
            self.image_pushButton.setEnabled(True)
            self.val_pushButton.setEnabled(True)
            self.hier_pushButton.setEnabled(True)
            self.format_comboBox.setCurrentIndex(0)
            self.name_lineEdit.clear()
            self.file_lineEdit.clear()
            self.idims_lineEdit.clear()
            self.odims_lineEdit.clear()
            self.label_lineEdit.clear()
            self.output_lineEdit.clear()
            self.image_lineEdit.clear()
            self.val_lineEdit.clear()
            self.hier_lineEdit.clear()
            self.padd_lineEdit.clear()
            self.pmul_lineEdit.clear()
            self.fp16_checkBox.setChecked(False)
            self.replace_checkBox.setChecked(False)
            self.verbose_checkBox.setChecked(False)
        else:
            modelName = self.upload_comboBox.currentText()
            setupDir = '~/.mivisionx-inference-analyzer'
            analyzerDir = os.path.expanduser(setupDir)
            for line in open(analyzerDir + "/setupFile.txt", "r"):
                tokens = line.split(';')
                if len(tokens) > 1:
                    name = tokens[1]
                    if modelName == name:
                        if tokens[0] == 'caffe':
                            format = 0
                        elif tokens[0] == 'onnx':
                            format = 1
                        else:
                            format = 2
                        self.format_comboBox.setCurrentIndex(format)
                        self.name_lineEdit.setText(tokens[1])
                        self.file_lineEdit.setText(tokens[2])
                        self.idims_lineEdit.setText(tokens[3])
                        self.odims_lineEdit.setText(tokens[4])
                        self.label_lineEdit.setText(tokens[5])
                        self.output_lineEdit.setText(tokens[6])
                        self.image_lineEdit.setText(tokens[7])
                        self.val_lineEdit.setText(tokens[8])
                        self.hier_lineEdit.setText(tokens[9])
                        self.padd_lineEdit.setText(tokens[10])
                        self.pmul_lineEdit.setText(tokens[11])
                        self.fp16_checkBox.setChecked(True) if tokens[12] == 'yes\n' or tokens[12] == 'yes' else self.fp16_checkBox.setChecked(False)
                        self.replace_checkBox.setChecked(True) if tokens[13] == 'yes\n' or tokens[13] == 'yes' else self.replace_checkBox.setChecked(False)
                        self.verbose_checkBox.setChecked(True) if tokens[14] == 'yes\n' or tokens[14] == 'yes' else self.verbose_checkBox.setChecked(False)
                        self.name_lineEdit.setEnabled(False)
                        self.file_lineEdit.setEnabled(False)
                        self.idims_lineEdit.setEnabled(False)
                        self.odims_lineEdit.setEnabled(False)
                        self.label_lineEdit.setEnabled(False)
                        self.output_lineEdit.setEnabled(False)
                        self.image_lineEdit.setEnabled(False)
                        self.val_lineEdit.setEnabled(False)
                        self.hier_lineEdit.setEnabled(False)
                        self.padd_lineEdit.setEnabled(False)
                        self.pmul_lineEdit.setEnabled(False)
                        self.fp16_checkBox.setEnabled(False)
                        self.output_pushButton.setEnabled(False)
                        self.label_pushButton.setEnabled(False)
                        self.image_pushButton.setEnabled(False)
                        self.val_pushButton.setEnabled(False)
                        self.hier_pushButton.setEnabled(False)
                        self.verbose_checkBox.setEnabled(False)
                        self.file_pushButton.setEnabled(False)
                        self.format_comboBox.setEnabled(False)

    def checkInput(self):
        if not (self.file_lineEdit.text() == '') and not (self.name_lineEdit.text() == '') \
            and not (self.idims_lineEdit.text() == '')  and not (self.odims_lineEdit.text() == '') \
            and not (self.output_lineEdit.text() == '') and not (self.label_lineEdit.text() == '') \
            and not (self.image_lineEdit.text() == ''):
                self.run_pushButton.setEnabled(True)
                self.run_pushButton.setStyleSheet("background-color: lightgreen")
        else:
            self.run_pushButton.setEnabled(False)
            self.run_pushButton.setStyleSheet("background-color: 0")

    def runConfig(self):
        self.model_format = self.format_comboBox.currentText()
        self.model_name = self.name_lineEdit.text()
        self.model = self.file_lineEdit.text()
        self.input_dims = '%s' % (self.idims_lineEdit.text())
        self.output_dims = '%s' % (self.odims_lineEdit.text())
        self.label = self.label_lineEdit.text()
        self.output = self.output_lineEdit.text()
        self.image = self.image_lineEdit.text()
        self.val = self.val_lineEdit.text()
        self.hier = self.hier_lineEdit.text()
        if len(self.padd_lineEdit.text()) < 1:
            self.add = '[0,0,0]'
        else:
            self.add = '[%s]' % (self.padd_lineEdit.text())
        if len(self.pmul_lineEdit.text()) < 1:
            self.multiply = '[1,1,1]'
        else:
            self.multiply = '[%s]' % (self.pmul_lineEdit.text())
        self.fp16 = 'yes' if self.fp16_checkBox.isChecked() else 'no'
        self.replace = 'yes' if self.replace_checkBox.isChecked() else 'no'
        self.verbose = 'yes' if self.verbose_checkBox.isChecked() else 'no'
        self.runningState = True
        self.close()

    def closeEvent(self, event):
        if not self.runningState:
            exit(0)
